/*
 * USBAEMon.java
 *
 * Created on February 17, 2005, 7:54 AM
 */

package ch.unizh.ini.caviar.usbaemonitor;

/**
 *  Acquires data from Anton Civit's Group's simple USB AER board that uses Silicon Labs (http://www.silabs.com)
 * C8051F320 controller and SiLabs USBXPress device and host driver firmware and software.
 *<p>
 * The DLL's USBAEMonitor.dll and SiUSBXp.dll must be accessible for windows programs.   Generally, this means they must be
 * somewhere on the PATH, for example, in <code>WINNT\system32</code>, or they can be in the directory the program is started.
 *It is generally simplest to just put the folders on the PATH variable.
 *<p>
 *Events are captured as 16 bit addresses and 32 bit timestamps with 1us tick. To use this class, construct an instance of <code>USBAEMonitor</code>,
 *then {@link #open} it. Each time you want to capture available events, call {@link #acquireAvailableEventsFromDriver}, which returns the number of
 *events grabbed. Access these events with {@link #getAddresses} and {@link #getTimestamps}, which each create new java arrays of the captured
 *data. {@link #overrunOccured} can be used to see if there was a driver overrun.
 *<p>
 *See the main() method for an example of use.
 *
 *
 * @author  tobi
 */
public class USBAEMonitor {
    
    static boolean libLoaded=false;
    
    static{
        try {
            //native init.  InterbusEcoLinkInterface.dll is built in InterbusEcoLinkInterface.
            System.loadLibrary("USBAEMonitor");// Load Library for interfacing to Eco-Link
            libLoaded=true;
            System.out.println("USBAEMonitor: loaded dynamic link library USBAEMonitor.dll");
        } catch (UnsatisfiedLinkError e) {
            //logging is special here because this one is static
            System.err.println("Can't load USBAEMonitor.dll");
            System.err.println("Not found in " + System.getProperty("java.ext.dirs")+" or "+System.getProperty("java.library.path"));
            throw new UnsatisfiedLinkError("USBAEMonitor.dll");
        }
        //        if(libLoaded){
        //            System.out.println("Registering static shutdown hook to close USBAEMonitor");
        //            Runtime.getRuntime().addShutdownHook(new Thread(){
        //                public void run(){ System.out.println("USBAEMonitor shutdown hook");}
        //            });
        //        }
        System.loadLibrary("USBAEMonitor");
        //        if(initIDs()!=0){ throw new USBAEMonitorException("can't init object field IDs");}
    }
    
    /** Creates a new instance of USBAEMonitor. Note that it is possible but probably bad to construct several instances
     and use each of them to open and read from the same device.
     This class would probably better be a singleton until multiple devices are supported.
     */
    public USBAEMonitor() {
    }
    
    //    private static native int initIDs();
    
    private native short[] nativeGetAddresses();
    private native int[] nativeGetTimestamps();
    private native boolean nativeOverrunOccured();
    private native int nativeOpen();
    private native int nativeClose();
    private native int nativeAcquireAvailableEventsFromDriver();
    private native int nativeGetNumEventsAcquired();
    
    /** Opens the device driver and gets a handle to the device which is internally
     * maintained. Timestamps are reset to 0 and buffers are flushed.
     *To reset timestamps, {@link #close} and reopen the device. It is presently an
     *error to open the device again after it has already been opened and not closed.
     *
     * @see #close
     * @return 0 if always.
     *@throws USBAEMonitorException if there is a problem. Diagnostics are printed to stdout in the native code.
     */
    public int open() throws USBAEMonitorException {
        int status=nativeOpen();
        if(status==0) return 0; else throw new USBAEMonitorException("nativeOpen");
    }
    
    /** Gets available events from driver. Call this before calling {@link #getAddresses }
     * or {@link #getTimestamps }. {@link USBAEMonitorException} is thrown if there is an error. 
     *{@link #overrunOccured} will be true if these was an overrun of the host USBXPress driver buffers (>16k events).
     *
     * @return number of events acquired. If this is zero there is no point in getting the events, because there are none.
     *@throws USBAEMonitorException
     * .
     */
    public int acquireAvailableEventsFromDriver() throws USBAEMonitorException {
        int numEvents=nativeAcquireAvailableEventsFromDriver();
        if (numEvents<0) throw new USBAEMonitorException("nativeAcquireAvailableEventsFromDriver");
        return numEvents;
    }
    
    /** Returns the number of events acquired by the last call to {@link
     * #acquireAvailableEventsFromDriver }
     * @return number of events acquired
     */
    public int getNumEventsAcquired(){
        return nativeGetNumEventsAcquired();
    }
    
    /** Returns a newly constructed array of addresses
     * @return array of addresses
     */
    public short[] getAddresses(){
        return nativeGetAddresses();
    }
    
    /** Returns a newly constructed array of timestamps that have a 1us tick
     * @return the timestamps
     */
    public int[] getTimestamps(){
        return nativeGetTimestamps();
    }
    
    /** Is true if an overrun occured in the driver (>16k events) the last time {@link
     * #acquireAvailableEventsFromDriver } was called. This flag is cleared by the next {@link #acquireAvailableEventsFromDriver}.
     *If there is an overrun, the events grabbed are the most ancient; events after the overrun are discarded. The timestamps continue on but will 
     *probably be lagged behind what they should be.
     * @return true if there was an overrun.
     */
    public boolean overrunOccured(){
        return nativeOverrunOccured();
    }
    
    /** Closes the device and frees the internal device handle. Never throws an exception.
     * @return 0 if successful
     */
    public int close(){
        return nativeClose();
    }
    
    
    /**
     * Tests event acquisition.
     *
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        int i;
        USBAEMonitor aemon=new USBAEMonitor();
        try{
            aemon.open();
        }catch(Exception e){
            e.printStackTrace();
            aemon.close();
            System.exit(1);
        }
        for(i=0;i<15;i++){
            int numEvents=0;
            try{
                numEvents=aemon.acquireAvailableEventsFromDriver();
            }catch(Exception e){
                e.printStackTrace();
                aemon.close();
                System.exit(1);
            }
            if(aemon.overrunOccured()){
                System.out.print("overrun ");
            }
            if(numEvents==0) System.out.println("no events");
            if(numEvents>0){
                System.out.print(numEvents+" events: ");
                short[] addr=aemon.getAddresses();
                int [] ts=aemon.getTimestamps();
                if(addr==null) continue; // didn't get anything
                int k=numEvents<10?numEvents:10;
                for(int j=0;j<k;j++){
                    short a=addr[j];
                    short x=(short)((a&0x7e)>>1); // map to retina address space
                    short y=(short)((a&0x3f00)>>8);
                    byte pol=(byte)((a&1)==1?1:-1);
                    System.out.print(x+","+y+","+ts[j]+" ");
                }
                System.out.println("");
            }
            try {Thread.currentThread().sleep(1000);} catch (java.lang.InterruptedException e) {}
        }
        aemon.close();
        System.out.println("closed");
        System.exit(0);
    }
}

