function showLiveLearning(arg)
%function showLiveLearning(arg)
%show live retina output using usbaemon, record events, etc
%just run without arguments to display retina output
% type h for help.
% by default uses openGL for rendering speed, g to toggle if there is a problem.
% a wireless network card can drastically reduce performance.
% author tobi feb 2005

persistent cs stopEnabled frameDuration colorMapping retinaSize
persistent usePolarity reconstructionEnabled reconstructionTau
persistent loggingEnabled allAddr allTs
persistent pauseEnabled numEventsCaptured timeCaptured frameDelay renderer jetMap grayColorCs
persistent fr im
persistent selectedPixel

if nargin==0,
    frameDuration=.01; % seconds to collect spikes per displayed frame
    cs=2; % color scale: this number of spikes produces saturated color
    colorMapping=0;  % 0 for gray, 1 to use RG for OFF/ON, 2 for time coherence
    usePolarity=0; % 1 to show spike polarity
    reconstructionEnabled=0; % set to 1 to reconstruct from events instead of displaying each 'frame' independently
    reconstructionTau=.2; % time constant of fading memory of past state of pixels
    retinaSizeX=8; % X size of retina (this is not general, just for plot limits)
    retinaSizeY=4; % Y size of retina (this is not general, just for plot limits)
    zscale=cs; % scale for gray level
    stopEnabled=0; %used to stop loop
    loggingEnabled=0; % used to flag event logging
    allAddr=[];
    allTs=[];
    numEventsCaptured=[]; timeCaptured=[];
    pauseEnabled=0;
    frameDelay=1/50; %1/100; % delay between each cycle
    renderer='OpenGL';
    jetMap=buildJetMap(cs);
    fr=zeros(retinaSizeY,retinaSizeX,3);
end

if nargin==1 && strcmp(arg,'stop'),
    stopEnabled=1;
end

if nargin==1 && strcmp(arg,'buttonup'),
    p=get(gca,'CurrentPoint');
    p=p(1,1:2);
    if p(1)<-1 || p(1)>retinaSizeY+1 || p(2)<-1 || p(2)>retinaSizeX+1,
        selectedPixel=[];
        fprintf('cleared pixel selection\n');
    else
        p=round(p);
        p(1)=max(1,p(1)); p(1)=min(retinaSizeY,p(1));
        p(2)=max(1,p(2)); p(2)=min(retinaSizeX,p(2));
        selectedPixel=p-1;
        fprintf('selected pixel %d,%d\n',selectedPixel(1),selectedPixel(2));
    end
    return
elseif nargin==1 && strcmp(arg,'key'),
    k=get(gcf,'currentkey');
    switch k,
        case 'h'
            disp 'c cycle display color'
            disp 'p toggle polarity'
            disp '= increase contrast'
            disp '- decrease constrast'
            disp 'l toggle logging'
            disp 'space pause/resume'
            disp 'g cycle renderer'
            disp 'x exit'
        case 'g'
            switch renderer,
                case 'painters',
                    renderer='OpenGL'
                case 'OpenGL',
                    renderer='zbuffer'
                case 'zbuffer',
                    renderer='painters'
            end
            set(gcf,'Renderer',renderer);
        case 'c'
            colorMapping=colorMapping+1;
            if colorMapping>2, colorMapping=0; end;
            if colorMapping==0,
                cs=grayColorCs; % restore
            end
            if colorMapping==2, 
                grayColorCs=cs; % save for restoring it
                cs=63; % because cs=1 doesn't give anything sensible
                jetMap=buildJetMap(cs); 
            end; 
        case 'p'
            usePolarity=~usePolarity;
        case {'equal','add','uparrow','rightarrow'}
            cs=cs-1; if cs<1, cs=1; end
            jetMap=buildJetMap(cs);
        case {'hyphen','subtract','downarrow','leftarrow'}
            cs=cs+1; if cs>64, cs=64; end
            jetMap=buildJetMap(cs);
        case 'space'
            pauseEnabled=~pauseEnabled;
        case 'l'
            loggingEnabled=~loggingEnabled;
            if ~loggingEnabled, % logging ended
                logFilename=['events',datestr(now,30)];
                %                save(logFilename, 'allAddr', 'allTs');
                save(logFilename);
                fprintf('events saved to %s.mat\n',logFilename);
            else % logging started
                fprintf('started logging events\n');
                allAddr=[]; allTs=[];
            end
        case 'x'
            stopEnabled=1;
            title('stopping...','color','r');
            drawnow;
            return
    end
    return
end

figure(1);
set(gcf,'name','Retina');
clf;
set (gcf, 'DoubleBuffer','on','Renderer',renderer);
set (gcf, 'MenuBar', 'none');
set(gcf,'keypressfcn','showLiveLearning(''key'')')
set(gcf,'WindowButtonUpFcn','showLiveLearning(''buttonup'');');
bg=0;
set(gcf,'color',[bg,bg,bg]);
set(gcf,'closerequestfcn','showLiveLearning(''stop'');');
set(gcf,'backingstore','off');
im=image(fr);

clear usbaemon; % clear to flush buffers and reset
startNow=now;
while ~stopEnabled,
    if stopEnabled,
        break;
    end
    if pauseEnabled, 
        pause(.3); 
        continue; 
    end;
    try
        [addr,ts]=usbaemon;
    catch
        pause(0.5);
        title(lasterr,'color','r');
        drawnow;
        continue;
    end
    if loggingEnabled,
        numEventsCaptured=[numEventsCaptured length(addr)];
        timeCaptured=[timeCaptured now-startNow];
        allAddr=[allAddr, addr];
        allTs=[allTs,ts];
    end
    [x,y,pol]=extractLearningEventsFromAddr(addr);
    if 1 %length(addr)>0,
        try
        switch colorMapping,
            case 0,            % each addr sums into frame according to coloring choice
                fr=fr*0;
                for i=1:length(x)
                    if ((y(i)<retinaSizeY) & (x(i)<retinaSizeX))
                        fr(y(i)+1,x(i)+1,1)=fr(y(i)+1,x(i)+1,1)+1;
                    %playSpikeIfSelected(x(i),y(i),pol(i),selectedPixel);
                    end;
                end
                % events have summed +1 for on, -1 for off.
                % there are 2*cs+1 0-1 levels, black is 0, white is 1, and gray is .5. thus 0->.5 always
                fr=max(0,min(1,fr/(2*cs))); 
                fr(:,:,2)=fr(:,:,1);
                fr(:,:,3)=fr(:,:,1); %gray
                set(im,'CData',fr);
            case 1,
                fr=fr*0;
                p=pol/2+.5+1;
                for i=1:length(x),
                    if ((y(i)<retinaSizeY) & (x(i)<retinaSizeX))
                          fr(y(i)+1,x(i)+1,p(i))=fr(y(i)+1,x(i)+1,p(i))+1;
                    end;
                    %playSpikeIfSelected(x(i),y(i),pol(i),selectedPixel);
                end;
                fr=max(0,min(1,fr/cs));
                set(im,'CData',fr);
            case 2,
                fr=fr*0;
                %fr=zeros(retinaSize,retinaSize,3); % start at black
                t=double(ts); % timestamps
                mint=t(1); maxt=t(end);
                t=(t-mint)/(maxt-mint+eps); % t are relative times normalized during 'frame'
                t=fix(1+floor(cs*t));% index into colormap according to temporal offset in 'frame'
                for i=1:length(x),
                    if ((y(i)<retinaSizeY) & (x(i)<retinaSizeX))
                          fr(y(i)+1,x(i)+1,:)=jetMap(t(i),:);
                    %playSpikeIfSelected(x(i),y(i),pol(i),selectedPixel);
                    end;
                end
                set(im,'CData',fr);
        end
        catch % in case there is problem with sound card
            disp(lasterr);
            selectedPixel=[];
        end
    end
    %mesh(fr);
    set(gca,'zlim',[-zscale,zscale],'xlim',[.5,retinaSizeX+.5],'ylim',[.5,retinaSizeY+.5],'ydir','normal','xticklabel',[],'yticklabel',[]);
    axis square
    if length(addr)>0,
        dt=(double(ts(end))-double(ts(1)))/1e3; % dt in ms
        if length(addr)<2,
            rate=0;
        else
            rate=length(addr)/(dt+eps);
        end
        s=sprintf('dt=%.0f ms, %d evts (%.1f keps), FS=%d evts',dt,length(addr),rate,cs);
        if loggingEnabled,
            s=[s, ' logging'];
        end
        title(s,'color','g');
        % debug ts
        %         t=double(ts);
        %         dt=diff(t);
        %         jt=find(dt>2^16);
        %         for i=1:length(jt),
        %             if i>1 & i<length(jt),
        %                 fprintf('wrap error at event %d, timestamps %X,%X,%X\n',jt(i),t(jt(i-1)),t(jt(i)),t(jt(i+1)));
        %             end
        %         end
        % end debug
        %     else
        %         s=sprintf('No events');
    end
    drawnow; % force display
    pause(frameDelay);
    %if length(addr)<15 & length(addr)>7, break; end
end
cleanup
return

function cleanup
clear usbaemon
delete(figure(1));
return

function jm=buildJetMap(cs)
jm=[0 0 0;jet(cs)];
return

function playSpikeSound(pol)
persistent spikeSoundSamples spikeSoundSamplesOn spikeSoundSamplesOff
if isempty(spikeSoundSamples),
    nSamps=200;
    x=linspace(0,pi,nSamps);
    y=sin(x).^4;
    z=gradient(y,x);
    z=z/max(z);
    spikeSoundSamples=z;
    zz=zeros(nSamps,1);
    spikeSoundSamplesOn=[z',zz];
    spikeSoundSamplesOff=[zz,z'];
end
if nargin==0,
    sound(spikeSoundSamples,44100,8);
elseif pol==1,
    sound(spikeSoundSamplesOn,44100,8);
else
    sound(spikeSoundSamplesOff,44100,8);
end
return

function playSpikeIfSelected(x,y,pol,selectedPixel)
if ~isempty(selectedPixel),
    if x==selectedPixel(1) && y==selectedPixel(2),
        playSpikeSound(pol);
    end
end
return

