function playRecordedRetina(cmd,arg)
% plays back events recorded with usbaemon
% prompts for .mat file holding allAddr and allTs recorded by logging from
% showLiveRetina
%             c cycle display color mapping
%             p toggle polarity enabled
%             uparrow/+/= increase contrast
%             downarrow/- decrease constrast
%             space pause/resume
%             r rewind to start
%             f faster (longer frames)
%             s slower (shorter frames)
%             b toggle play backwards
%             %1,2 mark start and end points
%             g cycle renderer
%             m toggle start/stop recording avi Movie
%             l toggle Loop through start and end points
%             t Toggle text enabled (speeds up rendering)
%             o rOtate 90 deg. each press rotates another 90deg
%             e show mEan center of mass and error on plot
%             . single step
%             x exit
% type 'h' in figure for help

global retina

persistent filename frameDuration cs useColor usePolarity tstart tend
persistent stopEnabled pauseEnabled rewToStartEnabled rewEnabled ffEnabled
persistent backwardsEnabled rewindToStartEnabled
persistent playbackFigure timeSlider abststart abstend
persistent markStart markEnd loopEnabled renderer jetMap grayColorCs
persistent colorMapping im textEnabled numEvents sliderMoved recordMovieEnabled mov
persistent buttonDown sliderValue
persistent selectedPixel pixelIsSelected rotation
persistent showCOMEnabled COMTau comLine
persistent singleStepEnabled
persistent retinaSize
persistent totalDuration

tick=1e-6; % timestamp tick in seconds for recorded data

% customizeable variables. these can be changed for another retina of
% different size and address encoding
if isempty(retina)
    retina='tmpdiff64'
end

if nargin==0 | ( nargin==2 & strcmp(cmd,'playFile')),
    if nargin==0,
        [filename,path]=uigetfile('*.mat','Select recorded retina data file');
        if filename==0, return; end
        load([path,filename]);
        cd(path);
    else
        load(arg)
    end
    
    if exist('ae','var') && size(ae,1)>0,
        ae = double (ae);
    elseif exist('allAddr','var') && exist('allTs','var'),
        ae=[double(allAddr);double(allTs)];
    else
        error 'data in matrix ae or vectors allAddr,allTs not found in this data file'
    end
    numEvents=size(ae,2);
    abststart=double(ae(2,1)); % first spike
    abstend=ae(2,end);
    totalDuration=(abstend-abststart)/1e6;
    fprintf('total duration %f seconds\n',totalDuration);

    colorMapping=0;  % 0 for gray, 1 to use RG for OFF/ON, 2 for time coherence

    frameDuration=.01; % seconds to collect spikes per displayed frame
    cs=1; % color scale: this number of spikes produces saturated color
    useColor=0;  % 0 for BW, 1 for RG, 2 for time=color
    usePolarity=1; % 1 to show spike polarity
    backwardsEnabled=0;
    rewindToStartEnabled=0;
    loopEnabled=1;
    pauseEnabled=0;
    markStart=abststart;
    markEnd=abstend;
    stopEnabled=0;
    renderer='OpenGL';
    jetMap=buildJetMap(cs);
    textEnabled=1;
    sliderMoved=0;
    recordMovieEnabled=0;
    buttonDown=0;
    rotation=0;
    showCOMEnabled=0;
    comLine=[];
    COMTau=.0001; % time const of COM averaging
    singleStepEnabled=0;
    
    % init figure
    if isempty(playbackFigure), playbackFigure=10; end;
    playbackFigure=figure;
    set (gcf, 'DoubleBuffer','on','Renderer',renderer);
    %set (gcf, 'MenuBar', 'none');
    set(gcf,'KeyPressFcn','playRecordedRetina(''key'');');
    set(gcf,'WindowButtonDownFcn','playRecordedRetina(''buttondown'');');
    set(gcf,'WindowButtonUpFcn','playRecordedRetina(''buttonup'');');
    set(gcf,'DeleteFcn','playRecordedRetina(''stop'');');
    timeSlider=uicontrol(playbackFigure,'Style','slider','position',[20,20,300,20]);
    set(timeSlider,'Callback','playRecordedRetina(''sliderMoved'');');
    %set(timeSlider,'ButtonDownFcn','playRecordedRetina(''buttondown'');');
    im=image;
    % we must initialize eventExtractor here, after file is loaded, because
    % eventExtractor may already exist in the data file as [] (before this was
    % fixed to clear it. it could not be serialized, that's why it was in the
    % file.
    if strcmp(retina,'tmpdiff64'),
        retinaSize=64; % size of retina (this is not general, just for plot limits)
        retinaEventExtractorFunction=@extractRetinaEventsFromAddr; %'extractRetinaEventsFromAddr';  % this is function that returns x,y,pol from AE, change for different retina, e.g. tmpdiff128
    elseif strcmp(retina,'tmpdiff128'),
        retinaSize=128; % size of retina (this is not general, just for plot limits)
        retinaEventExtractorFunction=@extractRetina128EventsFromAddr; %'extractRetinaEventsFromAddr';  % this is function that returns x,y,pol from AE, change for different retina, e.g. tmpdiff128
    end
    eventExtractor=[];
    try
        eventExtractor=getEventExtractor(retina);  % get the java class that extracts x,y,pol from ae (this speeds up things a lot)
    catch
        warning(['couldn''t get java event extractor for ' retina ', will be slightly slower to render']);
    end


end
    
if nargin==1 & strcmp(cmd,'buttondown'),
    %disp buttondown
    buttonDown=1;
    return
elseif nargin==1 & strcmp(cmd,'buttonup'),
    p=get(gca,'CurrentPoint');
    p=p(1,1:2);
    if p(1)<-1 || p(1)>retinaSize+1 || p(2)<-1 || p(2)>retinaSize+1,
        selectedPixel=[];
        fprintf('cleared pixel selection\n');
        pixelIsSelected=0;
    else
        p=round(p);
        p(1)=max(1,p(1)); p(1)=min(retinaSize,p(1));
        p(2)=max(1,p(2)); p(2)=min(retinaSize,p(2));
        selectedPixel=p-1;
        fprintf('selected pixel %d,%d\n',selectedPixel(1),selectedPixel(2));
        pixelIsSelected=1;
    end
    return
elseif nargin==1 & strcmp(cmd,'key'),
    k=get(gcf,'currentkey');
    switch k,
        case 'h'
            helpwin playRecordedRetina
        case 'c'
            colorMapping=colorMapping+1;
            if colorMapping>3, colorMapping=0; end;
            if colorMapping==0,
                cs=grayColorCs; % restore
            end
            if colorMapping==2,
                grayColorCs=cs; % save for restoring it
                cs=63; % because cs=1 doesn't give anything sensible
                jetMap=buildJetMap(cs);
            end;
        case 'g'
            switch renderer,
                case 'painters',
                    renderer='OpenGL'
                case 'OpenGL',
                    renderer='zbuffer'
                case 'zbuffer',
                    renderer='painters'
            end
            set(gcf,'Renderer',renderer);
        case 'p'
            usePolarity=~usePolarity;
        case {'equal','add','uparrow'} % more contrast
            cs=cs-1; if cs<1, cs=1; end
            jetMap=buildJetMap(cs);
        case {'hyphen','subtract','downarrow'}
            cs=cs+1; if cs>64, cs=64; end
            jetMap=buildJetMap(cs);
        case 's' % slower (shorter frames)
            frameDuration=frameDuration/2;
            if frameDuration<1e-5, frameDuration=1e-5; end
        case 'f' % faster (longer frames)
            frameDuration=frameDuration*2;
            if frameDuration>10, frameDuration=10; end
        case 'b' % toggle direction
            backwardsEnabled=~backwardsEnabled;
        case 'space'
            pauseEnabled=~pauseEnabled;
        case 'r'
            rewindToStartEnabled=1;
            %         case '1'
            %             markStart=tstart;
        case 'm'
            recordMovieEnabled=~recordMovieEnabled;
            if recordMovieEnabled,
                mov=avifile('retina.avi');
            else,
                mov=close(mov);
            end

            %         case '2'
            %             markEnd=tend;
        case 'l'
            loopEnabled=~loopEnabled;
        case 't'
            textEnabled=~textEnabled;
            if ~textEnabled,
                title('');
            end
        case 'x'
            stopEnabled=1;
            return
        case 'o'
            rotation=rotation+1; 
            if rotation>3, rotation=0; end
        case 'e'
            showCOMEnabled=~showCOMEnabled;
            if ~showCOMEnabled && ~isempty(comLine),
                delete(comLine);
                comLine=[];
            end
        case 'period'
            singleStepEnabled=~singleStepEnabled;
    end
    return
elseif nargin==1 & strcmp(cmd,'stop'),
    stopEnabled=1;
    return
elseif nargin==1 & strcmp(cmd,'sliderMoved'),
    sliderMoved=1;
    sliderValue=get(timeSlider,'value');
    return
end

% we make time windows, and for each spike in the window, we decode the x
% and y addresses and polarity and add it to an empty 64x64 matrix that
% gets plotted with a fixed brightness scaling.

tstart=abststart;
tend=tstart+frameDuration/tick;
startin=1;
endin=1;

while ~stopEnabled,
    %double(tend)<double(ae(2,end));
    if stopEnabled,
        break;
    end

    if sliderMoved,
        sliderMoved=0;
        backwardsEnabled=0;
        v=sliderValue; % get(timeSlider,'value');
%         fprintf('slider value=%f\n',v);
        tstart=abststart+(abstend-abststart)*v;
        tend=tstart+frameDuration/tick;
        in=find(ae(2,:)<=tstart);
        if isempty(in), in=numEvents; end
        startin=in(end);
        endin=startin;
        continue
    elseif ~buttonDown,
        v=(tstart-abststart)/(abstend-abststart);
        if v<0, v=0; elseif v>1, v=1; end
        set(timeSlider,'value',v);
    end

    if rewindToStartEnabled,
        backwardsEnabled=0;
        tstart=abststart;
        tend=tstart+frameDuration/tick;
        startin=1;
        endin=1;
        rewindToStartEnabled=0;
        continue
    end

    if pauseEnabled,
        v=(tstart-abststart)/(abstend-abststart);
        if v<0, v=0; elseif v>1, v=1; end
        if ~stopEnabled,  set(timeSlider,'value',v); end
        drawnow;
        pause(.1);
%         continue;  % now rerender during pause, in case color changes
    end
    
    if ~pauseEnabled && tstart>=abstend %| tstart>markEnd,
        title 'at end';
        drawnow
        pause(.3);
        if loopEnabled,
            tstart=abststart;
            tend=tstart+frameDuration/tick;
            startin=1;
            endin=1;
        else
            continue;
        end
    end
    if ~pauseEnabled && tstart<abststart %| tstart<markStart,
        title 'at start';
        drawnow
        pause(.3);
        if loopEnabled,
            backwardEnabled=0; % go forwards from start
            tstart=abststart;
            tend=tstart+frameDuration/tick;
            startin=1;
            endin=1;
        else
            continue
        end
    end

    %mask aer addresses to ON and OFF address-strings
    % find spikes in frame window, using pre-sorted timestamps

    if ~pauseEnabled,
        if ~backwardsEnabled,
            while endin<numEvents && ae(2,endin)<tend,
                endin=endin+1;
            end
            addr=ae(1,startin:endin);
            ts=(ae(2,startin:endin));
            startin=endin+1;
        else
            while endin>1 && ae(2,endin)>tstart,
                endin=endin-1;
            end
            addr=ae(1,endin:startin);
            ts=(ae(2,endin:startin));
            startin=endin;
        end

        %         [x,y,pol]=retinaEventExtractorFunction(addr(sampleIndex)); % only extract the sampled events
        eventExtractor.extract(addr);
        x=eventExtractor.x;
        y=eventExtractor.y;
        pol=eventExtractor.pol;
    end
    
    switch colorMapping,
        case 0,            % each addr sums into frame according to coloring choice
            fr=zeros(retinaSize,retinaSize,3);
            for i=1:length(x),
                fr(y(i)+1,x(i)+1,:)=fr(y(i)+1,x(i)+1,:)+pol(i);
                if ~isempty(selectedPixel),
                    if x(i)==selectedPixel(1) && y(i)==selectedPixel(2),
                        playSpikeSound;
                    end
                end
            end
            % events have summed +1 for on, -1 for off.
            % there are 2*cs+1 0-1 levels, black is 0, white is 1, and gray is .5. thus 0->.5 always
            fr=max(0,min(1,.5+fr/(2*cs)));
        case 1,
            fr=zeros(retinaSize,retinaSize,3);
            p=pol/2+.5+1;
            for i=1:length(x),
                fr(y(i)+1,x(i)+1,p(i))=fr(y(i)+1,x(i)+1,p(i))+1;
            end
            fr=max(0,min(1,fr/cs));
        case 2,
            fr=zeros(retinaSize,retinaSize,3); % start at black
            if length(ts)>0, % to make sure to correctly deal with no events
                t=double(ts); % timestamps
                mint=t(1); maxt=t(end);
                t=(t-mint)/(maxt-mint+eps); % t are relative times normalized during 'frame'
                t=fix(1+floor(cs*t));% index into colormap according to temporal offset in 'frame'
                t=max(1,t); % fix funny negative t if there are some
                for i=1:length(x),
                    fr(y(i)+1,x(i)+1,:)=jetMap(t(i),:);
                end
            end
        case 3,  % grayscale time mapping
            fr=zeros(retinaSize,retinaSize,3); % start at black
            if length(ts)>0, % to make sure to correctly deal with no events
                %fr=zeros(retinaSize,retinaSize,3); % start at black
                t=double(ts); % timestamps
                mint=t(1); maxt=t(end);
                t=(t-mint)/(maxt-mint+eps); % t are relative times normalized during 'frame' 0 is oldest, 1 is youngest
                for i=1:length(x),
                    fr(y(i)+1,x(i)+1,:)=t(i);  % set grayscale
                end
            end
    end
    if rotation>0,
        fr(:,:,1)=rot90(fr(:,:,1),rotation);
        fr(:,:,2)=rot90(fr(:,:,2),rotation);
        fr(:,:,3)=rot90(fr(:,:,3),rotation);
    end
    set(im,'CData',fr); %image(fr);
    %mesh(fr);
    set(gca,'zlim',[-cs,cs],'xlim',[1,retinaSize],'ylim',[1,retinaSize],'ydir','normal','xticklabel',[],'yticklabel',[]);
    axis square

    if showCOMEnabled && ~isempty(x),
        % compute mixing factor using chosen time constant of COM filter
        alpha=frameDuration/COMTau;
        alpha=min(alpha,1);
        [xmean,ymean,xsigma,ysigma]=updateCOM(x,y,pol,ts,alpha);
        comLine=paintCOM(xmean,ymean,xsigma,ysigma,comLine);
    end
    
    if textEnabled,
        if ~backwardsEnabled,
            dirString=' Fwd';
        else
            dirString=' Back';
        end
        tnow=(tstart-abststart)*tick;
        n=length(addr);
        if frameDuration<1e-3,
            s=sprintf('%.0fus@%.3f/%.1fs, %d evts, %.0fkeps, FS=%d evts, %s',frameDuration*1e6,tnow,totalDuration,n,n/frameDuration/1e3,cs,dirString);
        else
            s=sprintf('%.0fms@%.3f/%.1fs, %d evts, %.0fkeps, FS=%d evts, %s',frameDuration*1e3,tnow,totalDuration,n,n/frameDuration/1e3,cs,dirString);
        end
        title(s);
    end
    drawnow; % force display

    if recordMovieEnabled,
        F=getframe(gcf);
        mov=addframe(mov,F);
    end

    %pause
    if ~pauseEnabled,
        % next frames times
        if ~backwardsEnabled,
            tstart=tstart+frameDuration/tick;
            tend=tstart+frameDuration/tick;
        else
            tstart=tstart-frameDuration/tick;
            tend=tstart+frameDuration/tick; % plays frames backward, but each frame is still forwards
        end
    end
    
    if singleStepEnabled,
        pause(.1);
    end

end % main loop
cleanup(playbackFigure);
return

function cleanup(figure)
delete(figure);
return

function jm=buildJetMap(cs)
jm=[0 0 0;jet(cs)];
return

% gives new mean and std dev for mean x and y given set of events and
% mixing fraction alpha, which is fraction of new event to use for each new
% estimate update
function [xmean,ymean,xsigma,ysigma]=updateCOM(x,y,pol,t,alpha)
persistent xm ym
if isempty(xm),
    xm=31; ym=31;  % assumes size
end
if isempty(x),
    xmean=xm;
    ymean=ym;
    xsigma=0;
    ysigma=0;
    return
end
xmean=(1-alpha)*xm+alpha*median(x);
ymean=(1-alpha)*ym+alpha*median(y);
xm=xmean;
ym=ymean;
xsigma=std(x);
ysigma=std(y);
return

function comLine=paintCOM(xmean,ymean,xsigma,ysigma,comLine)
if isempty(comLine),
    comLine=line;
    %comCenter=fill;
end
xsigma=3;
ysigma=3;  % override actual errors to make fixed-size box
set(comLine,'xdata',[xmean-xsigma,xmean+xsigma,xmean+xsigma,xmean-xsigma,xmean-xsigma]);
set(comLine,'ydata',[ymean-ysigma,ymean-ysigma,ymean+ysigma,ymean+ysigma,ymean-ysigma]);
set(comLine,'color',[0,0,1],'linewidth',2);
return


